/*
 * _globals.h
 *
 *  Created on: 15. aug. 2018
 *      Author: teig
 */

#ifndef GLOBALS_H_
    #define GLOBALS_H_

    #ifdef GLOBALS_H_ // To show that the below may also be defined in library space

        // BOOLEAN #include <stdbool.h> if C99
        // See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#bool
        typedef enum {false,true} bool; // 0,1 This typedef matches any integer-type type like long, int, unsigned, char, bool

        #define min(a,b) (((a)<(b))?(a):(b))
        #define max(a,b) (((a)>(b))?(a):(b))
        #define abs(a)   (((a)<0)?(-(a)):(a))

        #define t_swap(type,a,b) {type t = a; a = b; b = t;}

        #define NUM_ELEMENTS(array) (sizeof(array) / sizeof(array[0])) // Kernighan & Pike p22

        typedef signed int time32_t; // signed int (=signed) or unsigned int (=unsigned) both ok, as long as they are monotoneously increasing
                                     // XC/XMOS 100 MHz increment every 10 ns for max 2exp32 = 4294967296,
                                     // ie. divide by 100 mill = 42.9.. seconds

        typedef enum {led_on, led_off} led_on_low_e; // 0 is led_on

        #define AFTER_32(a,b) ((a-b)>0)

        #ifdef DO_ASSERTS
            #define ASSERT_DELAY_32(d) do {if (d > INT_MAX) fail("Overflow");} while (0) // Needs <so646.h<, <limits.h> and <xassert.h>
            // INT_MAX is 2147483647 is what fits into 31 bits or last value before a signed 32 bits wraps around
        #else
            #define ASSERT_DELAY_32(d)
        #endif

        #define NOW_32(tmr,time) do {tmr :> time;} while(0) // A tick is 10ns
        // Programming XC on XMOS Devices (Douglas Watt)
        //     If the delay between the two input values fits in 31 bits, timerafter is guaranteed to behave correctly,
        //     otherwise it may behave incorrectly due to overlow or underflow. This means that a timer can be used to
        //     measure up to a total of 2exp31 / (100 mill) = 21s.

    #endif

    #define IS_MYTARGET_VOID               0
    #define IS_MYTARGET_STARTKIT           1
    #define IS_MYTARGET_XCORE_200_EXPLORER 2
    #define IS_MYTARGET_XCORE_XA_MODULE    3 // Not able to flash it
    #define IS_MYTARGET_MIC_ARRAY          4

    #if (MYTARGET==MIC_ARRAY)
        #define IS_MYTARGET IS_MYTARGET_MIC_ARRAY
        #define NUM_LEDS       13
        #define NUM_LEDS_08_12 (NUM_LEDS-8)
    #else
        #error NO TARGET DEFINED
    #endif

    #ifndef DISPLAY_FAST_DARK
        #error define DISPLAY_FAST_DARK in makefile
    #else
        #if (DISPLAY_FAST_DARK>1)
            #error not defined
        #endif
    #endif

    #define DEBUG_PRINT_GLOBAL_APP 1 // 0: all printf off
                                     // 1: controlled locally in each xc file. Unit must be connected to xTIMEcomposer. Not for off-line battery usage!
                                     //    DESTROYS ALL TIMING! PRINT FROM ONE CORE DESTROYS FOR ALL CORES!

    #ifndef USE_XSCOPE
        #error USE_XSCOPE not defined in makefile
    #elif (USE_XSCOPE == 0)
        #define DO_XSCOPE_INT(name,val)
    #elif (USE_XSCOPE == 1)
        #if (DEBUG_PRINT_GLOBAL_APP == 0)
            // Compare (compare 0220 and 0236)
            #warning XSCOPE probably not working. Set to 1, plus some DO_DEBUG_PRINT must also be set
            // See https://www.teigfam.net/oyvind/home/technology/219-my-beep-brrr-notes/#xscope
        #endif
        // This takes about 0.2 us, but ROOT_DELAY_UNTIL_NEXT_ROUND_US must have some real
        // temp_degC for data to be corrected (XScope would say "missing data" if it's too fast)

        #define DO_XSCOPE_INT(name,val) xscope_int(name,val)

        // In "config.xscope" DISCRETE, which must be handled in offline mode, see
        // "xTIMEcomposer User Guide for tools version 14.0.x" page 133:
        // "Only continous probes can be displayed real-time."
        // If DISCRETE not used here, temps_degC are interpolated (saw tooth curve is not much of temp_degC)
        // Log file is generated as "xscope.xmt" and opens after I stop the RUN
        //
        // xScope scaling, if not top and bottom is not seen:
        #define XSCOPE_SCALE_MAX   1100
        #define XSCOPE_VALUE_TRUE  1000
        #define XSCOPE_VALUE_FALSE  100
        #define XSCOPE_SCALE_MIN      0

        #define DELAY_FIRST_PRINT_MS 2500 // 2000 seen to fail. 1500 not ok for XSCOPE to work "automagically". Update: 48 kHz problems, 8 kHz ok!
    #else
        #error
    #endif

    // Requires #include <iso646.h> (not tested)
    #define BIT_HIGH(val,iofbit) do { val = val bitor       (1<<iofbit); } while (0)
    #define BIT_LOW(val,iofbit)  do { val = val bitand compl(1<<iofbit); } while (0)
    #define BIT_INV(val,iofbit)  do { val = val xor         (1<<iofbit); } while (0)

    #ifndef DO_DEBUG_1PORTS
        #error DO_DEBUG_1PORTS not defined in makefile
    #elif (DO_DEBUG_1PORTS == 1)
        #define DEBUG_1PORT_NUM_PULSE_CYCLES 10 // 1 barely visible on my scope
        #define DEBUG_1PORT_HIGH(p)        do { p <: 1; } while (0)
        #define DEBUG_1PORT_LOW(p)         do { p <: 0; } while (0)
        #define DEBUG_1PORT_HIGH_PULSE(p)  do { for (unsigned ix=0; ix<DEBUG_1PORT_NUM_PULSE_CYCLES; ix++) { p <: 1; } p <: 0; } while (0)
        #define DEBUG_1PORT_LOW_PULSE(p)   do { for (unsigned ix=0; ix<DEBUG_1PORT_NUM_PULSE_CYCLES; ix++) { p <: 0; } p <: 1; } while (0)
        #define DEBUG_1PORT_INVERT(p,val)  do { val = not val; p <: val; } while (0)
    #elif (DO_DEBUG_1PORTS == 0)
        #define DEBUG_1PORT_HIGH(p)
        #define DEBUG_1PORT_LOW(p)
        #define DEBUG_1PORT_HIGH_PULSE(p)
        #define DEBUG_1PORT_LOW_PULSE(p)
    #else
        #error
    #endif

    #define TEST_COUNTED_ARRAY 0 // 0 or 1
#else
    #error NESTED .h FILE
#endif /* GLOBALS_H_ */
